import { KasPublicKeyInfo, OriginAllowList } from '../access.js';
import { AuthProvider } from '../auth/providers.js';
export interface ClientConfig {
    allowedKases?: string[];
    ignoreAllowList?: boolean;
    authProvider: AuthProvider;
    dpopEnabled?: boolean;
    dpopKeys?: Promise<CryptoKeyPair>;
    ephemeralKeyPair?: Promise<CryptoKeyPair>;
    kasEndpoint: string;
}
/**
 * A Client encapsulates sessions interacting with TDF3 and nanoTDF backends, KAS and any
 * plugin-based sessions like identity and further attribute control. Most importantly, it is responsible
 * for local key and token management, including the ephemeral public/private keypairs
 * used for encrypting and decrypting information.
 *
 * @link https://developer.mozilla.org/en-US/docs/Web/API/CryptoKeyPair
 *
 * @example
 * import { Client, clientAuthProvider, decrypt, encrypt } from '@opentdf/sdk/nanotdf`
 *
 * const OIDC_ENDPOINT = 'http://localhost:65432/auth/';
 * const KAS_URL = 'http://localhost:65432/kas';
 *
 * let client = new Client(
 *    await clientAuthProvider({
 *      clientId: 'tdf-client',
 *      clientSecret: '123-456',
 *      oidcOrigin: OIDC_ENDPOINT,
 *    }),
 *    KAS_URL
 *  );
 *
 * // t=1
 * let nanoTDFEncrypted = await encrypt('some string', client.unwrappedKey);
 * let nanoTDFDecrypted = await decrypt(nanoTDFEncrypted, client.unwrappedKey);
 * nanoTDFDecrypted.toString() // 'some string'
 *
 */
export default class Client {
    static readonly KEY_ACCESS_REMOTE = "remote";
    static readonly KAS_PROTOCOL = "kas";
    static readonly SDK_INITIAL_RELEASE = "0.0.0";
    static readonly INITIAL_RELEASE_IV_SIZE = 3;
    static readonly IV_SIZE = 12;
    allowedKases: OriginAllowList;
    protected kasUrl: string;
    kasPubKey?: KasPublicKeyInfo;
    readonly authProvider: AuthProvider;
    readonly dpopEnabled: boolean;
    dissems: string[];
    dataAttributes: string[];
    protected ephemeralKeyPair: Promise<CryptoKeyPair>;
    protected requestSignerKeyPair: Promise<CryptoKeyPair>;
    protected iv?: number;
    /**
     * Create new NanoTDF Client
     *
     * The Ephemeral Key Pair can either be provided or will be generate when fetching the entity object. Once set it
     * cannot be changed. If a new ephemeral key is desired it a new client should be initialized.
     * There is no performance impact for creating a new client IFF the ephemeral key pair is provided.
     */
    constructor(optsOrOldAuthProvider: AuthProvider | ClientConfig, kasUrl?: string, ephemeralKeyPair?: CryptoKeyPair, dpopEnabled?: boolean);
    /**
     * Add attribute to the TDF file/data
     *
     * @param attribute The attribute that decides the access control of the TDF.
     */
    addAttribute(attribute: string): void;
    /**
     * Rewrap key
     *
     * @important the `fetchEntityObject` method must be called prior to
     * @param nanoTdfHeader the full header for the nanotdf
     * @param kasRewrapUrl key access server's rewrap endpoint
     * @param magicNumberVersion nanotdf container version
     * @param clientVersion version of the client, as SemVer
     */
    rewrapKey(nanoTdfHeader: ArrayBufferLike, kasRewrapUrl: string, magicNumberVersion: ArrayBufferLike, clientVersion: string): Promise<CryptoKey>;
}
//# sourceMappingURL=Client.d.ts.map